/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.util.binary;

import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

/**
 * Rancom access memory blob
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class BinaryMemory implements RandomAccessBinary {

	ByteOrder order = ByteOrder.LITTLE_ENDIAN;
	ByteBuffer buf;
	long pointer;
	/** the number of bytes added spare to the buffer when it is incremented */
	int increment;
	
	/**
	 * New memory blob
	 */
	public BinaryMemory() {
		buf = ByteBuffer.allocate(16);
		buf.order( order );
		increment = 16;
	}

	/**
	 * New memory blob
	 */
	public BinaryMemory(byte[] data) {
		this.buf = ByteBuffer.wrap(data);
		this.buf.order( ByteOrder.LITTLE_ENDIAN );
		this.order = ByteOrder.LITTLE_ENDIAN;
	}
	
	
	/**
	 * New memory blob
	 */
	public BinaryMemory(int initialSize) {
		buf = ByteBuffer.allocate(initialSize);
		buf.order( order ); 
		increment = Math.max(16, initialSize);
	}
	
	/**
	 * New memory blob
	 */
	public BinaryMemory(int initialSize, int increment) {
		buf = ByteBuffer.allocate(initialSize);
		buf.order( order ); 
		this.increment = increment;
	}
	
	/**
	 * Assume an existing byte buffer
	 * 
	 * @param buf buffer
	 */
	public BinaryMemory(ByteBuffer buf) {
		this.buf = buf;
		this.order = buf.order();
	}
	
	/**
	 * Get the backend byte buffer. The result may change if BinaryMemory is 
	 * written.
	 * 
	 * @return byte buffer
	 */
	public ByteBuffer toByteBuffer() {
		return buf;
	}
	
	@Override
	public void close() throws IOException {
	}

	@Override
	public void flush() throws IOException {
	}

	@Override
	public byte get() {
		assertHasReadableBytes(1);		
		buf.position( (int) pointer );
		byte result = buf.get();
		pointer += 1;
		return result;
	}

	@Override
	public void get(byte[] dst, int offset, int length) {
		assertHasReadableBytes(length);
		buf.position( (int) pointer );
		buf.get(dst, offset, length);
		pointer += length;
	}

	@Override
	public void get(byte[] dst) {
		assertHasReadableBytes(dst.length);		
		buf.position( (int) pointer );
		buf.get(dst);
		pointer += dst.length;
	}

	@Override
	public void get(ByteBuffer buf) {		
		int bytes = buf.remaining();
		assertHasReadableBytes( bytes );
		buf.position( (int) pointer );
		if (buf.hasArray()) {
			this.buf.get(buf.array(), buf.arrayOffset() + buf.position(), bytes);
			buf.position(buf.capacity());
		} else {
			buf.put(buf);
		}
		pointer += bytes;
	}

	@Override
	public void get(ByteBuffer buf, int length) {
		assertHasReadableBytes(length);
		buf.position( (int) pointer );
		if (buf.hasArray()) {
			this.buf.get(buf.array(), buf.arrayOffset() + buf.position(), length);
			buf.position(buf.position() + length);
		} else {
//			int len = Math.min( Math.min( buf.remaining(), this.buf.remaining() ), length);
			int len = length;
			int origLimit = this.buf.limit();
			try {
				this.buf.limit(this.buf.position()+len);
				buf.put(this.buf);
			} finally {
				this.buf.limit(origLimit);
			}
		}
		pointer += length;
	}

	@Override
	public double getDouble() {
		assertHasReadableBytes(8);
		buf.position( (int) pointer );
		double value = buf.getDouble();
		pointer += 8;
		return value;
	}

	@Override
	public float getFloat() {
		assertHasReadableBytes(4);
		buf.position( (int) pointer );
		float result = buf.getFloat();
		pointer += 4;
		return result;
	}

	@Override
	public int getInt() {
		assertHasReadableBytes(4);
		buf.position( (int) pointer );
		int value = buf.getInt();
		pointer += 4;
		return value;
	}

	@Override
	public long getLong() {
		assertHasReadableBytes(8);
		buf.position( (int) pointer );
		long value = buf.getLong();
		pointer += 8;
		return value;
	}

	@Override
	public short getShort() {
		assertHasReadableBytes(2);
		buf.position( (int) pointer );
		short result = buf.getShort();
		pointer += 2;
		return result;
	}

	@Override
	public long length() {
		return buf.limit();
	}
	
	@Override
	public long position() {
		return pointer;
	}
	
	@Override
	public void position(long newPosition) throws IOException {
		this.pointer = newPosition;
	}
	
	public ByteOrder order() {
		return buf.order();
	}

	public void order(ByteOrder order) {
		this.order = order;
		buf.order(order);
	}

	@Override
	public void put(byte b) throws IOException {
		assertHasWritableBytes(1);
		buf.position( (int) pointer );
		buf.put(b);
		pointer += 1;
	}

	@Override
	public void put(ByteBuffer src) throws IOException {
		long bytes = src.remaining();
		assertHasWritableBytes( bytes );
		buf.position( (int) pointer );
		buf.put(src);
		pointer += bytes;
	}

	@Override
	public void put(ByteBuffer src, int length) throws IOException {
		assertHasWritableBytes(length);
		buf.position( (int) pointer );
		if (src.hasArray()) {
			byte[] array = src.array();
			buf.put(array, src.arrayOffset() + src.position(), length);
		} else {
			for (int i=0; i<length; i++)
				buf.put(src.get());
		}
		pointer += length;
	}

	
	public void put(InputStream is) throws IOException {
		long oldLen = length();
		while (is.available()>0) {
			int n = is.available();
			assertHasWritableBytes(n);
			byte buf[] = this.buf.array();
			n = is.read(buf, (int) pointer, n);			
			pointer += n;
		}
		long newLen = length();
		if (newLen>oldLen & pointer<newLen) setLength(pointer);
	}

	@Override
	public void put(byte[] src, int offset, int length) throws IOException {
		assertHasWritableBytes(length);
		buf.position( (int) pointer );
		buf.put(src, offset, length);
		pointer += length;
	}

	@Override
	public void put(byte[] src) throws IOException {
		assertHasWritableBytes(src.length);
		buf.position( (int) pointer );
		buf.put(src);
		pointer += src.length;
	}

	@Override
	public void putDouble(double value) throws IOException {
		assertHasWritableBytes(8);
		buf.position( (int) pointer );
		buf.putDouble(value);
		pointer += 8;
	}

	@Override
	public void putFloat(float value) throws IOException {
		assertHasWritableBytes(4);
		buf.position( (int) pointer );
		buf.putFloat(value);
		pointer += 4;
	}

	@Override
	public void putInt(int value) throws IOException {
		assertHasWritableBytes(4);
		buf.position( (int) pointer );
		buf.putInt(value);
		pointer += 4;
	}

	@Override
	public void putLong(long value) throws IOException {
		assertHasWritableBytes(8);
		buf.position( (int) pointer );
		buf.putLong(value);
		pointer += 8;
	}

	@Override
	public void putShort(short value) throws IOException {
		assertHasWritableBytes(2);
		buf.position( (int) pointer );
		buf.putShort(value);
		pointer += 2;
	}

	void assertHasReadableBytes(long count) {
		if (pointer + count > buf.limit())
			throw new IndexOutOfBoundsException();
	}
	
	void assertHasWritableBytes(long count) 
	throws IOException {
		if (pointer + count > buf.limit()) 
			setLength(pointer+count);
	}


	@Override
	public void insertBytes(long bytes) throws IOException {
		if (pointer<0) throw new IndexOutOfBoundsException();
		// 1. Increase length
		long oldLength = length();
		long newLength = Math.max(oldLength + bytes, pointer + bytes);
		// No need to allocate
		if (newLength<buf.capacity()) {
			// Update limit
			buf.limit((int)newLength);
			
			// No need to move bytes
			if (pointer>oldLength) return;
				
			// Move bytes
			if (buf.hasArray()) {
				System.arraycopy(buf.array(), (int) pointer, buf.array(), (int) (pointer+bytes), (int) (oldLength - pointer));
			} 
			else 
			{
				byte[] b = new byte[(int) (oldLength - pointer)];
				buf.position((int)pointer);
				buf.get(b);
				buf.position( (int) (pointer + bytes) );
				buf.put(b);
			}
		} else
		// Need to reallocate
		{
			long inc = Math.max(increment, newLength / 4);		
			ByteBuffer oldBuf = buf;
			ByteBuffer newBuf = ByteBuffer.allocate((int) (newLength + inc));
			newBuf.order(order);
			newBuf.limit((int)newLength);
			
			newBuf.position(0);
			oldBuf.position(0);

			// Copy left side
			oldBuf.get(newBuf.array(), 0, (int)pointer);
			
			// Copy right side
			if (pointer<oldLength) 
			oldBuf.get(newBuf.array(), (int)(pointer+bytes), (int)(oldLength-pointer));
			
			buf = newBuf;
		}
		
	}
	
	@Override
	public void removeBytes(long bytes) throws IOException {
		long oldLength = length();
		long newLength = oldLength - bytes;
		if (pointer<0 || pointer>=length()) throw new IndexOutOfBoundsException();
		
		// move some bytes
		if (buf.hasArray()) {
			System.arraycopy(buf.array(), (int)(pointer+bytes), buf.array(), (int) pointer, (int) (newLength-pointer) ); 
		} else {
			byte[] b = new byte[(int) (oldLength-pointer)];
			buf.position((int) (pointer+bytes));
			buf.get(b);
			buf.position((int) (pointer));
			buf.put(b);
		}
		
		buf.limit( (int) newLength );
	}

	@Override
	public void setLength(long newLength) throws IOException {
		int oldLength = (int) length();
		if (oldLength==newLength) return;
		if (newLength<=oldLength) {
			buf.limit((int)newLength);
		} else
		if (newLength<=buf.capacity()) {
			buf.limit((int)newLength);
		} else {
			ByteBuffer oldBuf = buf;
			long inc = Math.max(increment, newLength / 4);
			ByteBuffer newBuf = ByteBuffer.allocate((int) (newLength + inc));
			newBuf.order(order);
			oldBuf.position(0);
			oldBuf.get(newBuf.array(), 0, oldLength);
			newBuf.limit((int)newLength);
			buf = newBuf;
		}		
	}

	@Override
	public void skip(long bytes) throws IOException {
		pointer += bytes;
	}

	@Override
	public String toString() {
		return "Mem("+length()+")";
	}
	
}

