/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.util.binary;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

/**
 * This class is a Random Access File implementation to BinaryWritable
 * implementation. The implementation is buffered, therefore the values
 * may not be written until {@link #flush()}. The implementation is not 
 * multi-thread safe. Although file pointer of the backend RandomAccessFile 
 * is allowed to move between read operations of this class.
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class FileWriteable implements BinaryWriteable, Seekable {

	RandomAccessFile file;
	ByteOrder order = ByteOrder.LITTLE_ENDIAN;
	
	/** The internal write buffer */
	byte buf[] = new byte[8192];
	
	/** The number of valid bytes in the buffer */
	int count;
	
	/** The position where to flush the buffer in file. The value is valid only when count > 0 */
	long filePosition;
	
	/**
	 * Create a BinaryWritable implemantation wrapped over a RandomAccessFile.
	 * 
	 * @param file
	 * @throws IOException 
	 */
	public FileWriteable(RandomAccessFile file) throws IOException
	{
		this.file = file;
		filePosition = file.getFilePointer();
	}
	
	public void close() throws IOException {
		try {
			flush();
		} catch (IOException ignored) {			
		}
	}
	
	public RandomAccessFile getFile() {
		return file;
	}
	
	@Override
	public ByteOrder order() {
		return order;
	}

	@Override
	public void order(ByteOrder order) {
		if (order==null) throw new IllegalArgumentException("null argument");
		this.order = order;
	}

	void _put(int value) throws IOException
	{
		if (buf==null) {
			file.writeByte(value);			
			filePosition++;
		} else {
			if (count >= buf.length) flush();			
			buf[count++] = (byte)value;
		}
	}
	
	void _put(byte[] src, int offset, int length) throws IOException {		
		if (buf == null || length >= buf.length) {
			file.write(src, offset, length);
			filePosition += length;
			return;
		}
		
		if (length > buf.length - count) {
			flush();
		}
		
		System.arraycopy(src, offset, buf, count, length);
		count += length;
	}
	
	@Override
	public void put(byte b) throws IOException {
		_put(b);
	}

	@Override
	public void put(ByteBuffer src) throws IOException {
		if (src.hasArray()) {
			byte array[] = src.array();
			_put(array, src.position(), src.remaining());
			src.position(src.limit()); 
		} else 
			for (;src.hasRemaining();)
				_put(src.get());
	}

	@Override
	public void put(ByteBuffer src, int length) throws IOException {
		if (src.hasArray()) {
			byte array[] = src.array();
			_put(array, src.position(), length);
			src.position(length); 
		} else {
			for (int i=0; i<length; i++)
				_put(src.get());
		}
	}

	@Override
	public void put(byte[] src, int offset, int length) throws IOException {
		_put(src, offset, length);
	}

	@Override
	public void put(byte[] src) throws IOException {
		_put(src, 0, src.length);
	}

	@Override
	public void putDouble(double value) throws IOException {
		putLong(Double.doubleToLongBits(value));
	}

	@Override
	public void putFloat(float value) throws IOException {
		putInt(Float.floatToIntBits(value));
	}

	@Override
	public void putInt(int value) throws IOException {
		if (order == ByteOrder.BIG_ENDIAN) {
			_put(value >> 24);
			_put(value >> 16);
			_put(value >> 8);
			_put(value);
		} else {
			_put(value);
			_put(value >> 8);
			_put(value >> 16);
			_put(value >> 24);
		}
	}

	@Override
	public void putLong(long value) throws IOException {
		if (order == ByteOrder.BIG_ENDIAN) {
			_put((int) (value >> 56));
			_put((int) (value >> 48));
			_put((int) (value >> 40));
			_put((int) (value >> 32));
			_put((int) (value >> 24));
			_put((int) (value >> 16));
			_put((int) (value >> 8));
			_put((int) (value));
		} else {
			_put((int) (value));
			_put((int) (value >> 8));
			_put((int) (value >> 16));
			_put((int) (value >> 24));
			_put((int) (value >> 32));
			_put((int) (value >> 40));
			_put((int) (value >> 48));
			_put((int) (value >> 56));
		}
	}

	@Override
	public void putShort(short value) throws IOException {
		if (order == ByteOrder.BIG_ENDIAN) {
			_put(value >> 8);
			_put(value);
		} else {
			_put(value);
			_put(value >> 8);
		}
	}

	@Override
	public void flush() throws IOException 
	{
		if (buf==null || count==0) return;
		
		file.seek(filePosition);
		file.write(buf, 0, count);			
		filePosition += count;
		count = 0;
	}
	
	public long position() {
		return filePosition + count;
	}
	
	public void position(long newPosition) 
	throws IOException 
	{
		if (count>0 && newPosition!=filePosition+count)
			flush();
		filePosition = newPosition;		
	}

}

